"""Python project helpers."""

import tsh
import click

from recept.apps import rm, find, black, flake8, pydocstyle, pytest


def clean(path: str):
    """Cleanup the project directory from python artifacts.

    Args:
        path: Path to the project directory.
    """
    with tsh.pushd(path):
        # Delete the __pycache__ directories
        r = find(".", "-type", "d", "-name", "__pycache__")
        for path in r.stdout.decode("utf-8").splitlines():
            rm(path)

        # Delete *.py[cod] files if some is left
        r = find(".", "-type", "f", "-name", "*.py[cod]")
        for path in r.stdout.decode("utf-8").splitlines():
            rm(path)

        # Delete jupyter checkpoints
        rm(".ipynb_checkpoints")

        # Remove py.test artifacts
        rm(".pytest_cache")


def shell(namespace=None):
    """Run IPython shell with loaded variables."""
    try:
        from IPython import embed

        user_ns = {}
        if namespace is not None:
            user_ns.update(namespace)
        embed(user_ns=user_ns, colors="neutral")
    except ImportError:
        click.secho("IPython is not installed", fg="red")


def fmt(path: str, target_version: str = "py37", line_length: int = 79):
    """Format the code in path using black.

    Args:
        path: Path to the project dir.
        target_version: Python target version py35, py36, py37...
        line_length: Desired line length.
    """
    black(
        "--safe",
        "--target-version",
        target_version,
        "--line-length",
        str(line_length),
        path,
    )


def check(path: str):
    """Run code checks using flake8 and pydocstyle.

    Args:
        path: Path to the project directory.
    """
    flake8(path)
    pydocstyle(path)


def test(path: str):
    """Run tests using pytest.

    Args:
        path: Path to the project directory.
    """
    pytest("-v", "--cov", path, path)



import click


class ReceptError(click.ClickException):
    pass
    # TODO: Revert when we replace click.
    # def __init__(self, message):
    #     self.message = message
    #
    # @property
    # def name(self):
    #     return self.__class__.__name__
    #
    # def __str__(self):
    #     return f"{self.name}({self.message})"
    #
    # __repr__ = __str__


class ImproperlyConfigured(ReceptError):
    pass



from typing import List, Dict, Optional, Union

import click

from recept.apps import docker


def login(username: str, password: str, registry: str):
    """Login to docker registry.

    Args:
        username: Username.
        password: Password.
        registry: Url to the docker registry.
    """
    return docker.login(
        "--username", username, "--password", password, registry
    )


def remove_dangling_images():
    r = docker.images("-f", "dangling=true", "-q", _out=None)
    images = r.stdout.decode("utf-8").split()
    if len(images) > 0:
        docker.rmi(*images)


def build(
    dockerfile: str,
    image: str,
    tag: str = "latest",
    latest: bool = True,
    build_dir: str = ".",
):
    """Build docker container.

    Args:
        dockerfile: Path to the Dockerfile.
        image: Image name.
        tag: Build tag.
        latest: Tag the image with the latest tag too. (If the tag is not
            already "latest").
        build_dir: Build directory.

    """

    docker.build("--file", dockerfile, "--tag", f"{image}:{tag}", build_dir)
    if latest and tag != "latest":
        docker.tag(f"{image}:{tag}", f"{image}:latest")


def pull(image: str, tag: str = "latest", registry: Optional[str] = None):
    """Pull the app image from ECR registry.

    Args:
        image: Image name.
        tag: Image tag.
        registry: Url to the registry. Default: docker hub.
    """
    docker.pull(
        f"{image}:{tag}" if registry is None else f"{registry}/{image}:{tag}"
    )


def push(image: str, tag: str = "latest", registry: Optional[str] = None):
    """Push app docker containers to registry.

    Args:
        image: Image name.
        tag: Image tag.
        registry: Url to the registry. Default: docker hub.
    """
    if registry is not None:
        # Tag the images with registry
        docker.tag(f"{image}:{tag}", f"{registry}/{image}:{tag}")
        image = f"{registry}/{image}:{tag}"
    else:
        image = f"{image}:{tag}"

    # Push images to registry
    docker.push(image)


def is_running(name: str):
    """Check if a container is running.

    Args:
        name: Name of the running container.
    """
    r = docker.ps("--filter", f"name={name}", "--format", "{{.Names}}")
    return f"{name}" == r.stdout.decode("utf-8").strip()


def create_network(network: str):
    """Create docker network if it doesn't already exist.

    Args:
        network: Network name.
    """
    r = docker.network.ls(
        "--filter", f"name={network}", "--format", "{{.Name}}"
    )
    if r.stdout.decode("utf-8").strip() != network:
        # Network does not exist, it should be created
        docker.network.create(network)


def start(
    name: str,
    image: str = None,
    tag: str = "latest",
    env: Optional[Dict[str, str]] = None,
    ports: Optional[Dict[int, int]] = None,
    volumes: Optional[Dict[str, str]] = None,
    networks: Optional[List[str]] = None,
    command: Optional[Union[str, List[str]]] = None,
    entrypoint: Optional[str] = None,
    interactive: bool = False,
):
    """Start a container.

    Args:
        name: Name of the container when started.
        image: Image to run. If image is not provided then:
            {registry}/{prefix}/{project} image will be used.
        tag: Image tag: Default: latest.
        env: Dictionary of environment variables that should be available to
            the container.
        ports: Dictionary or host_port -> guest_port mappings that should be
            exposed.
        volumes: Dictionary of host_path -> guest_path mappings that should be
            attached as volumes.
        networks: List of network names that the container should have access.
        command: Override of the default entrypoint of the container.
        entrypoint: Override of the default entrypoint.
        interactive: Should the container be started in interactive mode. If
            False, the container (or command) will be started in detached mode.
    """
    if is_running(name):
        click.secho(f"{name} is already running.", fg="red")
        return

    if interactive:
        docker_app = docker.bake(_fg=True)
        run_command = ["--interactive", "--tty", "--rm", "--name", name]
    else:
        docker_app = docker
        run_command = ["--detach", "--tty", "--rm", "--name", name]

    if entrypoint is not None:
        run_command.extend(["--entrypoint", entrypoint])

    # Add env
    if env is not None:
        for key, value in env.items():
            run_command.extend(["--env", f"{key}={value}"])

    # Add ports
    if ports is not None:
        for key, value in ports.items():
            run_command.extend(["--publish", f"{key}:{value}"])

    # Add volumes
    if volumes is not None:
        for key, value in volumes.items():
            run_command.extend(["--volume", f"{key}:{value}"])

    # Add networks
    if networks is not None:
        for network in networks:
            create_network(network)
            run_command.extend(["--network", network])

    # Add image
    run_command.append(f"{image}:{tag}")

    if command is None:
        docker_app.run(*run_command)
    else:
        if isinstance(command, str):
            command = [command]
        docker_app.run(*run_command, *command)


def stop(name: str):
    """Stop a running container.

    Args:
        name: Name of the running container.
    """
    if not is_running(name):
        click.secho(f"{name} is not running.", fg="red")
        return
    docker.stop(name)


def attach(name: str, command="sh"):
    """Attach to a running container.

    Args:
        name: Name of the running container.
        command: Command to run.
    """
    docker.bake(_fg=True).exec("--interactive", "--tty", name, command)



"""
Env allows you to use 12factor inspired env variables for configuration.

ToEnvMixin is a class that does the opposite - Creates an environment variable
dictionary from class attributes.
"""

__all__ = ["Env", "ToEnvMixin"]

import os
import re
import json
import logging
from pathlib import Path
from copy import deepcopy
from typing import Callable, Any, Optional, Dict
from urllib.parse import urlparse, ParseResult as UrlParseResult

from recept.errors import ImproperlyConfigured


logger = logging.getLogger(__name__)


BOOLEAN_TRUE_STRINGS = ("true", "on", "yes")


def parse_value(value: str, cast) -> Any:
    """Parse and cast provided value.

    Args:
        value: Stringed value.
        cast: Callable used for casting or type.

    Returns:
        Casted value.
    """
    if cast is None:
        return value
    elif cast is bool:
        return value.lower() in BOOLEAN_TRUE_STRINGS
    elif isinstance(cast, list):
        return list(map(cast[0], [x for x in value.split(",") if x]))
    elif isinstance(cast, tuple):
        val = value.strip("(").strip(")").split(",")
        return tuple(map(cast[0], [x for x in val if x]))
    elif isinstance(cast, dict):
        key_cast = cast.get("key", str)
        value_cast = cast.get("value", str)
        value_cast_by_key = cast.get("cast", dict())
        return dict(
            map(
                lambda kv: (
                    key_cast(kv[0]),
                    parse_value(
                        kv[1], value_cast_by_key.get(kv[0], value_cast)
                    ),
                ),
                [val.split("=") for val in value.split(";") if val],
            )
        )
    elif cast is dict:
        return dict(
            val.split("=", maxsplit=1) for val in value.split(",") if val
        )
    elif cast is list:
        return [x for x in value.split(",") if x]
    elif cast is tuple:
        val = value.strip("(").strip(")").split(",")
        return tuple([x for x in val if x])
    else:
        value = cast(value)
    return value


class NoValue(object):
    def __repr__(self):
        return "<{0}>".format(self.__class__.__name__)


class Env(object):
    """Provide scheme-based lookups of environment variables.

    This elevates the caller of doing conversion or passing default value.

    Usage::

        env = Env(RECEPT_DEBUG=bool, MONGO_USERNAME=(str, 'DEFAULT'))
        if env('RECEPT_DEBUG'):
            ...

    Args:
        prefix: Prefix used for variables. For example if prefix is set to
            `mongo` and you search for variable `username` then the env object
            will search for the key `MONGO_USERNAME` in the `os.environment`.
            If `prefix` is `None`, search key will only be `USERNAME`.s

        **scheme: Types definition for variables.
    """

    ENVIRON = os.environ
    NOTSET = NoValue()
    URL_CLASS = UrlParseResult

    def __init__(self, prefix: str = None, **scheme):
        self.prefix = prefix
        self.scheme = scheme

    def __str__(self):
        if self.prefix:
            return f"Env(prefix={self.prefix})"
        return "Env()"

    __repr__ = __str__

    def bake(self, prefix: str = None, **scheme):
        """Return a new instance of Env with merged parameters."""
        s = deepcopy(self.scheme)
        s.update(scheme)
        return Env(prefix=prefix, **s)

    def __call__(self, var, cast=None, default=NOTSET, parse_default=False):
        return self.get_value(
            var, cast=cast, default=default, parse_default=parse_default
        )

    def __contains__(self, var):
        if self.prefix is not None:
            var = f"{self.prefix}_{var}"
        return var.upper() in self.ENVIRON

    # Shortcuts

    def str(self, var, default=NOTSET, multiline=False) -> str:
        value = self.get_value(var, default=default)
        if multiline:
            return value.replace("\\n", "\n")
        return value

    def bytes(self, var, default=NOTSET, encoding="utf8") -> bytes:
        return self.get_value(var, default=default, cast=str).encode(encoding)

    def bool(self, var, default=NOTSET) -> bool:
        return self.get_value(var, cast=bool, default=default)

    def int(self, var, default=NOTSET) -> int:
        return self.get_value(var, cast=int, default=default)

    def float(self, var, default=NOTSET) -> float:
        return self.get_value(var, cast=float, default=default)

    def json(self, var, default=NOTSET) -> dict:
        return self.get_value(var, cast=json.loads, default=default)

    def list(self, var, cast=None, default=NOTSET) -> list:
        return self.get_value(
            var, cast=list if cast is None else [cast], default=default
        )

    def tuple(self, var, cast=None, default=NOTSET) -> tuple:
        return self.get_value(
            var, cast=tuple if cast is None else (cast,), default=default
        )

    def dict(self, var, cast=dict, default=NOTSET) -> dict:
        return self.get_value(var, cast=cast, default=default)

    def url(self, var, default=NOTSET) -> UrlParseResult:
        return self.get_value(
            var, cast=urlparse, default=default, parse_default=True
        )

    def path(self, var, default=NOTSET, **kwargs) -> Path:
        return Path(self.get_value(var, default=default), **kwargs)

    def get_value(
        self,
        var: str,
        cast: Optional[Callable[[str], Any]] = None,
        default: Any = NOTSET,
        parse_default: bool = False,
    ) -> Any:
        """Return value for given environment variable.

        Args:
            var: Name of variable.
            cast: Callable used to cast return value.
            default: If var not present in environ, return this instead.
            parse_default: Parse the default value too.

        Returns:
            Value from environment or default if value is not set.
        """

        logger.debug(
            "get '%s' casted as '%s' with default '%s'", var, cast, default
        )

        if var in self.scheme:
            var_info = self.scheme[var]

            try:
                has_default = len(var_info) == 2
            except TypeError:
                has_default = False

            if has_default:
                if not cast:
                    cast = var_info[0]

                if default is self.NOTSET:
                    try:
                        default = var_info[1]
                    except IndexError:
                        pass
            else:
                if not cast:
                    cast = var_info

        search_var = (
            var.upper()
            if self.prefix is None
            else f"{self.prefix}_{var}".upper()
        )
        try:
            value = self.ENVIRON[search_var]
        except KeyError:
            if default is self.NOTSET:
                raise ImproperlyConfigured(
                    f"Set the {search_var} environment variable"
                )

            value = default

        if (
            cast is None
            and default is not None
            and not isinstance(default, NoValue)
        ):
            cast = type(default)

        if value != default or (parse_default and value):
            value = parse_value(value, cast)

        return value


class ToEnvMixin:
    """ToEnvMixin provides a class to env variables dictionary function.

    It will only convert simple types to environment variables:

        str, bytes, bool, int, float

    Usage::

        >>> @dataclass
        ... class Mongo(ToEnvMixin):
        ...     host: str = "127.0.0.1"
        ...     port: int = 27017
        ...
        >>> mongo = Mongo()
        >>> mongo.to_env()
        {
            "MONGO_HOST": "127.0.0.1",
            "MONGO_PORT": "27017"
        }
    """

    # If env_prefix is None a snake case version of a class name will be used.
    env_prefix = None

    # Environment variable attributes.
    # If it's set to None, all keys from __dict__ will be considered as env
    # attributes.
    env_attrs = None

    __FIRST_CAP_RE = re.compile(r"(.)([A-Z][a-z]+)")
    __ALL_CAP_RE = re.compile(r"([a-z0-9])([A-Z])")
    __FORBIDDEN_CHARS_RE = re.compile(r"[^\w]+")

    @classmethod
    def __types(cls):
        return str, bytes, bool, int, float, ToEnvMixin

    @classmethod
    def __camel_case_to_snake_case(cls, s: str) -> str:
        """Convert camel case to snake case."""
        s1 = cls.__FORBIDDEN_CHARS_RE.sub("_", s)
        s2 = cls.__FIRST_CAP_RE.sub(r"\1_\2", s1)
        return cls.__ALL_CAP_RE.sub(r"\1_\2", s2).lower()

    def __get_env(self, prefix: str, attr: str) -> Dict[str, str]:
        obj = getattr(self, attr)

        if isinstance(obj, ToEnvMixin):
            return obj.to_env()

        key = (
            attr.upper()
            if prefix.strip() == ""
            else f"{prefix}_{attr}".upper()
        )
        if isinstance(obj, str):
            return {key: obj}
        elif isinstance(obj, bool):
            return {key: "true" if obj else "false"}
        elif isinstance(obj, bytes):
            return {key: str(obj, encoding="utf-8")}
        else:
            return {key: str(obj)}

    def to_env(self) -> dict:
        """Returns this class variables in env format dictionary."""
        prefix = self.__camel_case_to_snake_case(
            self.__class__.__name__
            if self.__class__.env_prefix is None
            else self.__class__.env_prefix
        )

        if self.__class__.env_attrs is None:
            env_attrs = [
                key
                for key, value in self.__dict__.items()
                if isinstance(value, self.__types())
            ]
        else:
            env_attrs = self.__class__.env_attrs

        env = {}
        for attr in env_attrs:
            env.update(self.__get_env(prefix=prefix, attr=attr))
        return env



import sys

import sh


def app(name, *args, _out=sys.stdout, _err=sys.stderr, _tee=True, **kwargs):
    try:
        return sh.Command(name).bake(
            *args, _out=_out, _err=_err, _tee=_tee, **kwargs
        )
    except sh.CommandNotFound:
        return sh.Command(sys.executable).bake(
            "-c",
            (
                f"import sys; import click; click.secho('Command `{name}` "
                f"not found', fg='red'); sys.exit(1)"
            ),
        )


# Shell commands
ls = app("ls")
rm = app("rm", "-rf")
cp = app("cp", "-rf")
find = app("find", _out=None)
mount = app("mount")
umount = app("umount", "-f")


# Python commands
python = app(sys.executable)
pip = app("pip")
pytest = app("py.test", "-s", _tee=False, _ok_code=[0, 1, 2, 3, 4, 5])
black = app("black")
flake8 = app("flake8", _ok_code=[0, 1])
pydocstyle = app("pydocstyle", _ok_code=[0, 1])


# Docker
docker = app("docker")



import os
import sys
import argparse
import importlib


NO_RECIPE_MODULE_ERROR = """\
Module 'recipes' not found.

You can either:

    1. Enter the directory where the `recipes` module is located.
    2. Add it's parent folder to PYTHONPATH.
    3. Point to the recipes module using the `--recipes` argument.
"""


HELP = """\
rr [-h] [--recipes RECIPES] COMMAND

Error: {error}

Recipe runner.

optional arguments:
  -h, --help         show this help message and exit
  --recipes RECIPES  Path to the recipes module.
"""


def print_help_and_exit(error):
    print(HELP.format(error=error))
    sys.exit(1)


def create_parser():
    parser = argparse.ArgumentParser(description="Recipe runner.")
    parser.add_argument("--recipes")
    return parser


def main():
    parser = create_parser()
    ns, rest = parser.parse_known_args()
    if ns.recipes is not None:
        recipes = os.path.abspath(ns.recipes)
        # Check if the folder exists.
        if not os.path.isdir(recipes):
            print_help_and_exit(f"Recipes folder not found: {recipes}")
        # If it exists add it's parent folder to PYTHONPATH
        sys.path.insert(0, os.path.dirname(recipes))
        sys.argv[1:] = rest
    else:
        # Add current working directory to the pythonpath if it's not already
        # added.
        cwd = os.getcwd()
        if cwd not in sys.path:
            sys.path.insert(0, cwd)

    # Try to import recipes
    try:
        recipes = importlib.import_module("recipes")
        cli = getattr(recipes, "cli")
        cli()
    except ModuleNotFoundError as e:
        print_help_and_exit(
            NO_RECIPE_MODULE_ERROR if e.name == "recipes" else e.msg
        )


if __name__ == "__main__":
    main()



import io
import sys
import enum
from typing import Union

import click


class Color(enum.Enum):
    """Available colors in terminal."""

    black = "black"
    red = "red"
    green = "green"
    yellow = "yellow"
    blue = "blue"
    magenta = "magenta"
    cyan = "cyan"
    white = "white"
    bright_black = "bright_black"
    bright_red = "bright_red"
    bright_green = "bright_green"
    bright_yellow = "bright_yellow"
    bright_blue = "bright_blue"
    bright_magenta = "bright_magenta"
    bright_cyan = "bright_cyan"
    bright_white = "bright_white"


def echo(
    message: str = None,
    file: io.TextIOBase = None,
    nl: bool = True,
    err: bool = False,
    fg: Union[str, Color] = None,
    bg: Union[str, Color] = None,
    bold: bool = None,
    dim: bool = None,
    underline: bool = None,
    blink: bool = None,
):
    """Print a message plus a newline to the given file or stdout.

    Supported color names are: black, red, green, yellow, blue, magenta, cyan,
    white, bright_black, bright_red, bright_green, bright_yellow, bright_blue,
    bright_magenta, bright_cyan, bright_white.

    Args:
        message: Message to print.
        file: File to write to (defaults to `stdout`).
        err: if set to `True` the file defaults to `stderr` instead of
            `stdout`.
        nl: if set to `True` (the default) a newline is printed afterwards.
        fg: if provided this will become the message foreground color.
        bg: if provided this will become the message background color.
        bold: if provided this will enable or disable bold mode.
        dim: if provided this will enable or disable dim mode.
        underline: if provided this will enable or disable underline.
        blink: if provided this will enable or disable blinking.
    """

    click.secho(
        message=message,
        file=file,
        nl=nl,
        err=err,
        fg=fg.value if isinstance(fg, Color) else fg,
        bg=bg.value if isinstance(bg, Color) else bg,
        bold=bold,
        dim=dim,
        underline=underline,
        blink=blink,
    )


def exit(status: int = None, message: str = None):
    """Exit the application with optional error message.

    If the status is omitted or None, it defaults to zero (i.e., success).
    If the status is an integer, it will be used as the system exit status.
    If the message is provided it will be printed to stderr.

    Args:
        status: Exit status. None = 0, i.e. success. Every other integer is
            interpreted as failure.
        message: Optional message printed to stderr before exiting the
            application.
    """
    if message is not None:
        echo(message=message, err=True, fg=Color.red)
    sys.exit(0 if status is None else status)



import functools


def register(cli, func, *args, **kwargs):
    """Register a function for a command line interface.

    Args:
        cli: Command line interface for which we register the function.
        func: Function that we register.
        args, kwargs: Optional positional and keyword arguments that will be
            passed to the function when invoked.
    """
    name = func.__name__
    if len(args) > 0 or len(kwargs) > 0:
        callback = functools.wraps(func)(
            functools.partial(func, *args, **kwargs)
        )
    else:
        callback = func
    return cli.command(name)(callback)



import os
import json
import pytest
from pathlib import Path
from recept.env import Env, parse_value
from recept.errors import ImproperlyConfigured


OLD_ENVIRON = None


class TestData:
    url = "http://www.google.com/"
    json = dict(one="bar", two=2, three=33.44)
    dict = dict(foo="bar", test="on")
    path = "/home/dev"
    exported = "exported var"


def generate_data():
    return dict(
        STR_VAR="bar",
        MULTILINE_STR_VAR="foo\\nbar",
        INT_VAR="42",
        FLOAT_VAR="33.3",
        BOOL_TRUE_VAR="on",
        BOOL_TRUE_VAR2="True",
        BOOL_FALSE_VAR="0",
        BOOL_FALSE_VAR2="False",
        INT_LIST="42,33",
        INT_TUPLE="(42,33)",
        STR_LIST_WITH_SPACES=" foo,  bar",
        EMPTY_LIST="",
        DICT_VAR="foo=bar,test=on",
        URL_VAR=TestData.url,
        JSON_VAR=json.dumps(TestData.json),
        PATH_VAR=TestData.path,
        EXPORTED_VAR=TestData.exported,
    )


def assert_type_and_value(type_, expected, actual):
    assert isinstance(actual, type_)
    assert expected == expected


def setup_module(module):
    module.OLD_ENVIRON = os.environ
    os.environ = Env.ENVIRON = generate_data()


def teardown_module(module):
    os.environ = module.OLD_ENVIRON


@pytest.fixture
def env():
    return Env()


def test_not_present_with_default(env):
    assert 3 == env("not_present", default=3)


def test_not_present_without_default(env):
    with pytest.raises(ImproperlyConfigured):
        env("not_present")


def test_contains(env):
    assert "STR_VAR" in env
    assert "EMPTY_LIST" in env
    assert "I_AM_NOT_A_VAR" not in env


def test_str(env):
    assert_type_and_value(str, "bar", env("STR_VAR"))
    assert_type_and_value(str, "bar", env.str("STR_VAR"))
    assert_type_and_value(str, "foo\\nbar", env.str("MULTILINE_STR_VAR"))
    assert_type_and_value(
        str, "foo\nbar", env.str("MULTILINE_STR_VAR", multiline=True)
    )


def test_bytes(env):
    assert_type_and_value(bytes, b"bar", env.bytes("STR_VAR"))


def test_int(env):
    assert_type_and_value(int, 42, env("INT_VAR", cast=int))
    assert_type_and_value(int, 42, env.int("INT_VAR"))


def test_int_with_none_default(env):
    assert env("NOT_PRESENT_VAR", cast=int, default=None) is None


def test_float(env):
    assert_type_and_value(float, 33.3, env("FLOAT_VAR", cast=float))
    assert_type_and_value(float, 33.3, env.float("FLOAT_VAR"))


def test_bool_true(env):
    assert_type_and_value(bool, True, env("BOOL_TRUE_VAR", cast=bool))
    assert_type_and_value(bool, True, env("BOOL_TRUE_VAR2", cast=bool))
    assert_type_and_value(bool, True, env.bool("BOOL_TRUE_VAR"))


def test_bool_false(env):
    assert_type_and_value(bool, False, env("BOOL_FALSE_VAR", cast=bool))
    assert_type_and_value(bool, False, env("BOOL_FALSE_VAR2", cast=bool))
    assert_type_and_value(bool, False, env.bool("BOOL_FALSE_VAR"))


def test_int_list(env):
    assert_type_and_value(list, [42, 33], env("INT_LIST", cast=[int]))
    assert_type_and_value(list, [42, 33], env.list("INT_LIST", int))


def test_int_tuple(env):
    assert_type_and_value(tuple, (42, 33), env("INT_LIST", cast=(int,)))
    assert_type_and_value(tuple, (42, 33), env.tuple("INT_LIST", int))
    assert_type_and_value(tuple, ("42", "33"), env.tuple("INT_LIST"))


def test_str_list_with_spaces(env):
    assert_type_and_value(
        list, [" foo", "  bar"], env("STR_LIST_WITH_SPACES", cast=[str])
    )
    assert_type_and_value(
        list, [" foo", "  bar"], env.list("STR_LIST_WITH_SPACES")
    )


def test_empty_list(env):
    assert_type_and_value(list, [], env("EMPTY_LIST", cast=[int]))


def test_dict_value(env):
    assert_type_and_value(dict, TestData.dict, env.dict("DICT_VAR"))


def test_dict_parsing(env):
    assert parse_value("a=1", dict) == {"a": "1"}
    assert parse_value("a=1", dict(value=int)) == {"a": 1}
    assert parse_value("a=1,2,3", dict(value=[str])) == {"a": ["1", "2", "3"]}
    assert parse_value("a=1,2,3", dict(value=[int])) == {"a": [1, 2, 3]}
    assert parse_value(
        "a=1;b=1.1,2.2;c=3", dict(value=int, cast=dict(b=[float]))
    ) == {"a": 1, "b": [1.1, 2.2], "c": 3}

    assert parse_value(
        "a=uname;c=http://www.google.com;b=True",
        dict(value=str, cast=dict(b=bool)),
    ) == {"a": "uname", "c": "http://www.google.com", "b": True}


def test_url_value(env):
    url = env.url("URL_VAR")
    assert url.__class__ == env.URL_CLASS
    assert url.geturl() == TestData.url
    assert env.url("OTHER_URL", default=None) is None


def test_json_value(env):
    assert env.json("JSON_VAR") == TestData.json


def test_path(env):
    path = env.path("PATH_VAR")
    assert_type_and_value(Path, Path(TestData.path), path)


def test_smart_cast(env):
    assert env.get_value("STR_VAR", default="string") == "bar"
    assert env.get_value("BOOL_TRUE_VAR", default=True)
    assert not env.get_value("BOOL_FALSE_VAR", default=True)
    assert env.get_value("INT_VAR", default=1) == 42
    assert env.get_value("FLOAT_VAR", default=1.2) == 33.3


def test_exported(env):
    assert env("EXPORTED_VAR") == TestData.exported


def test_singleton_environ():
    config = generate_data()

    class MyEnv(Env):
        ENVIRON = config

    env = MyEnv()
    assert env.ENVIRON is config


def test_schema():
    env = Env(
        INT_VAR=int,
        NOT_PRESENT_VAR=(float, 33.3),
        STR_VAR=str,
        INT_LIST=[int],
        DEFAULT_LIST=([int], [2]),
    )

    assert_type_and_value(int, 42, env("INT_VAR"))
    assert_type_and_value(float, 33.3, env("NOT_PRESENT_VAR"))

    assert env("STR_VAR") == "bar"
    assert env("NOT_PRESENT2", default="foo") == "foo"

    assert_type_and_value(list, [42, 33], env("INT_LIST"))
    assert_type_and_value(list, [2], env("DEFAULT_LIST"))

    # Override schema in this one case
    assert_type_and_value(str, "42", env("INT_VAR", cast=str))



from dataclasses import dataclass

from recept.env import ToEnvMixin


def env_class(prefix=None):
    @dataclass
    class EnvClass(ToEnvMixin):
        env_prefix = prefix

        str_value: str = "string"
        bytes_value: str = b"bytes"
        true_value: bool = True
        false_value: bool = False
        int_value: int = 42
        float_value: float = 3.14

        @staticmethod
        def expected(prefix):
            return {
                f"{prefix}STR_VALUE": "string",
                f"{prefix}BYTES_VALUE": "bytes",
                f"{prefix}TRUE_VALUE": "true",
                f"{prefix}FALSE_VALUE": "false",
                f"{prefix}INT_VALUE": "42",
                f"{prefix}FLOAT_VALUE": "3.14",
            }

    return EnvClass()


def nested_env_class(prefix=None, nested_prefix=None):
    @dataclass
    class NestedEnvClass(ToEnvMixin):
        env_prefix = prefix

        some_value: str = "foo"
        env_class: ToEnvMixin = env_class(prefix=nested_prefix)

        @staticmethod
        def expected(self_prefix, nest_prefix):
            env = {f"{self_prefix}SOME_VALUE": "foo"}
            env.update(env_class().expected(prefix=nest_prefix))
            return env

    return NestedEnvClass()


def test_prefix():
    # None
    obj = env_class()
    assert obj.to_env() == obj.expected(prefix="ENV_CLASS_")

    # Empty
    obj = env_class(prefix="")
    assert obj.to_env() == obj.expected(prefix="")

    # Custom
    obj = env_class(prefix="foo-bar")
    assert obj.to_env() == obj.expected(prefix="FOO_BAR_")


def test_nested():
    # None, None
    obj = nested_env_class()
    assert obj.to_env() == obj.expected(
        self_prefix="NESTED_ENV_CLASS_", nest_prefix="ENV_CLASS_"
    )

    # None, Empty
    obj = nested_env_class(nested_prefix="")
    assert obj.to_env() == obj.expected(
        self_prefix="NESTED_ENV_CLASS_", nest_prefix=""
    )

    # None, Custom
    obj = nested_env_class(nested_prefix="foo bar")
    assert obj.to_env() == obj.expected(
        self_prefix="NESTED_ENV_CLASS_", nest_prefix="FOO_BAR_"
    )

    # Empty, None
    obj = nested_env_class(prefix="")
    assert obj.to_env() == obj.expected(
        self_prefix="", nest_prefix="ENV_CLASS_"
    )

    # Empty, Empty
    obj = nested_env_class(prefix="", nested_prefix="")
    assert obj.to_env() == obj.expected(self_prefix="", nest_prefix="")

    # Empty, Custom
    obj = nested_env_class(prefix="", nested_prefix="foo! bar")
    assert obj.to_env() == obj.expected(self_prefix="", nest_prefix="FOO_BAR_")

    # Custom, None
    obj = nested_env_class(prefix="foo-bar")
    assert obj.to_env() == obj.expected(
        self_prefix="FOO_BAR_", nest_prefix="ENV_CLASS_"
    )

    # Custom, Empty
    obj = nested_env_class(prefix="FIZZ  !  buzz", nested_prefix="")
    assert obj.to_env() == obj.expected(
        self_prefix="FIZZ_BUZZ_", nest_prefix=""
    )

    # Custom, Custom
    obj = nested_env_class(prefix="a", nested_prefix="b")
    assert obj.to_env() == obj.expected(self_prefix="A_", nest_prefix="B_")



import io
from setuptools import setup, find_packages


version = "0.0.8"
author = "Viktor Kerkez"
author_email = "alefnula@gmail.com"
url = "https://github.com/alefnula/recept"


setup(
    name="recept",
    version=version,
    author=author,
    author_email=author_email,
    maintainer=author,
    maintainer_email=author_email,
    url=url,
    description="Recipe runner and a recipes functions library.",
    long_description=io.open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    platforms=["Windows", "POSIX", "MacOS"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
    ],
    license="Apache-2.0",
    packages=find_packages(),
    install_requires=io.open("requirements.txt").read().splitlines(),
    entry_points="""
        [console_scripts]
        rr=recept.__main__:main
    """,
)



__all__ = [
    "register",
    "Config",
    "c",
    "color",
    "Color",
    "echo",
    "exit",
    "ReceptError",
    # Click functions
    "group",
    "pass_context",
    "pass_obj",
    "argument",
    "option",
    "Path",
]


import click

from recept.cli import register
from recept.config import Config
from recept.termui import Color, echo, exit
from recept.errors import ReceptError

c = color = Color


# Click functions

group = click.group
pass_context = click.pass_context
pass_obj = click.pass_obj
argument = click.argument
option = click.option
Path = click.Path



import io
import os
from configparser import ConfigParser
from typing import Dict, Any, Callable
from datetime import datetime, timezone


CONFIG_DIR = os.path.expanduser("~/.recept")

# Timestamp format used in configuration
TIMESTAMP_FORMAT = "%Y.%m.%dT%H:%M:%S"


# Converter that convert a value from configuration string
CONVERTERS_FROM_CFG: Dict[Any, Callable[[str], Any]] = {
    bool: lambda value: value.lower() in {"true", "yes", "on"},
    datetime: lambda value: datetime.strptime(
        value, TIMESTAMP_FORMAT
    ).astimezone(timezone.utc),
}


# Converters that convert a value to configuration string
CONVERTERS_TO_CFG: Dict[Any, Callable[[Any], str]] = {
    bool: lambda value: str(bool(value)).lower(),
    datetime: lambda value: value.astimezone(timezone.utc).strftime(
        TIMESTAMP_FORMAT
    ),
}


class Config:
    """Recipes configuration store.

    Args:
        project: Name of the project for which the configuration is loaded and
            saved.
    """

    def __init__(self, project: str):
        self.project = project
        # Load the configuration from ini file
        self._parser = ConfigParser()
        self.load()

    @property
    def config_file(self):
        return os.path.join(CONFIG_DIR, f"{self.project}.ini")

    def __str__(self):
        return f"Config(project={self.project})"

    __repr__ = __str__

    def set(self, section: str, key: str, value: Any, converter: Any = str):
        if section not in self._parser:
            self._parser[section] = {}
        converter = CONVERTERS_TO_CFG.get(converter, converter)
        self._parser[section][key] = converter(value)
        self.save()

    def get(
        self,
        section: str,
        key: str,
        default: Any = None,
        converter: Any = None,
    ):
        value = self._parser.get(section, key, fallback=default)
        if value is default or converter is None:
            return value

        converter = CONVERTERS_FROM_CFG.get(converter, converter)
        return converter(value)

    def load(self):
        if os.path.isfile(self.config_file):
            self._parser.read(self.config_file)

    def save(self):
        # Create configuration directory if it doesn't exist
        os.makedirs(CONFIG_DIR, exist_ok=True)
        with io.open(self.config_file, "w", encoding="utf-8") as f:
            self._parser.write(f)
